/***************************************************************************
                          openmosixmigmon.cpp  -  description
                             -------------------
    begin                : Son Jan  5 15:59:03 CET 2003
    copyright            : (C) 2003 by Matt
    email                : mosixview@t-online.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <qaccel.h>
#include "openmosixmigmon.h"

#include "helpicon.xpm"
#include "migmonback.xpm"
#include "whatis.xpm"
#include "white.xpm"


// constructor
OpenMosixmigmonApp::OpenMosixmigmonApp()
{

  setCaption(tr("openMosixmigmon " VERSION));
  this->setGeometry(100, 100, ewide, eheight);
  this->setFixedSize(ewide, eheight);
  this->setclusternodes(gethowmanynodes());

  updatenodes();
  QTimer *updatetimer = new QTimer( this, "intern" );
  connect( updatetimer, SIGNAL(timeout()), this, SLOT(redraw()) );
  updatetimer->start(10000);

}


// destructor
OpenMosixmigmonApp::~OpenMosixmigmonApp()
{
}




//############# helping functions #####################

int file_exist(string filename) {
  std::ifstream readfilename(filename.c_str());
  if (readfilename) {
  readfilename.close();
    return 1;
    } else {
    return 0;
  }
}

int string2int(const string s) {
  if (s.length() == 0) {
    return -1;
    } else {
    return(atoi(s.c_str()));
   }
}

string int2string(const int i) {
  char buf[12];
  sprintf(buf,"%d",i);
  return(string(buf));
}


string onlyfirstpart(string firstpart) {
string endstr;
string::iterator p = firstpart.begin();
	for (p=firstpart.begin(); p!=firstpart.end(); p++) {
		string tmpstr;
		tmpstr = *p;
		if(tmpstr==" ") {
			return endstr;
		}
		endstr = endstr + tmpstr;
	}
	return endstr;
}


//############# method implementations #####################


 // check if the cluster-config changed
bool OpenMosixmigmonApp::checknodes() {
	if (gethowmanynodes()==clusternodes) {
		return true;
		} else {
   	return false;
	}
}


// set the clusternodes value
void OpenMosixmigmonApp::setclusternodes(int nodes) {
	clusternodes=nodes;
}

// returns howmany nodes we are
int OpenMosixmigmonApp::gethowmanynodes() {
	struct dirent **namelist;
	int n;
	int howmanynodes=0;

	n = scandir(clusterdirectory, &namelist, 0, 0);
	if (n < 0) {
		return 0;
		} else {
		while(n--) {
			if(!strchr(namelist[n]->d_name, '.')) {
				howmanynodes++;
				free(namelist[n]);
			}
		}
		free(namelist);
	return howmanynodes;
	}	
}


// converts a cluster id to an hostname
string OpenMosixmigmonApp::getmynodeID() {
        // gather nodes id
        string which;
        char zeile[201];
        string nodename;
        string mynodeid;

        size_t remit;

        which = which + clustercommand + " whois";
        which = which + " > /tmp/nodes.tmp";
        system(which.c_str());
        if (file_exist("/tmp/nodes.tmp")) {
                std::ifstream readnodename("/tmp/nodes.tmp");
                if (readnodename) {
                        readnodename.getline(zeile, 200);
                        nodename.erase();
                        nodename= nodename+zeile;
                        readnodename.close();
                        remit=nodename.find('#', 0);
                        remit++;
                        mynodeid=nodename.erase(0, remit);

                        } else {
                        cout << "could not read from /tmp/nodes.tmp"  << endl;
                        return 0;
                }
        }
   return mynodeid;
}




void OpenMosixmigmonApp::updatenodes() {

  QPixmap help, backgr, whatisthis;
  whatisthis = QPixmap(whatis);
  help = QPixmap(helpicon);
  backgr = QPixmap(migmonback);

  struct dirent **namelist;
  int n;
  int howmany=0;
  double nodeangle;
  double nodeX, nodeY;
  double noderadialvalue;
  int nodes=1;
  double mainradialvalue=0;
  string myid;

  QFont font("helvetica");
  font.setPixelSize(10);

  // who am i
  myid = getmynodeID();

  // the main label
  QLabel_main= new QLabel(this,"main");
  QLabel_main->setGeometry(0, 0, int(eheight), int(ewide));
  QLabel_main->setMinimumSize(0,0);
  QLabel_main->setBackgroundPixmap(backgr);


  // the main node
  QLabel_tux= new nodewidget(QLabel_main, myid.c_str());
  QLabel_tux->setGeometry(int(xcenter)-10,int(ycenter)-10,20,20);
  QLabel_tux->setMinimumSize(0,0);
  QLabel_tux->setPixmap(help);
  QLabel_tux->setFrameStyle( QFrame::Panel | QFrame::Sunken );


  QLabel_myid= new QLabel(QLabel_main,"tuxid");
  QLabel_myid->setGeometry(int(xcenter)-10,int(ycenter)+10,40,20);
  QLabel_myid->setMinimumSize(0,0);
  QLabel_myid->setFont(font);
  QLabel_myid->setText("");
  QLabel_myid->setFrameStyle( QFrame::Panel | QFrame::Sunken );


  helpme = new QToolButton(QLabel_main, "whatisthis");
  helpme->setGeometry(ewide-30, 5,25,25);
  helpme->setPixmap(whatisthis);
  connect(helpme, SIGNAL(clicked()), this, SLOT(slotHelp()));


  // empty nodelist
  nodelist.clear();

  // fill nodelist
  n = scandir(clusterdirectory, &namelist, 0, 0);
  if (n < 0) {
    perror("scandir");
    } else {
    while(n--) {
      if(!strchr(namelist[n]->d_name, '.')) {
      // add node id to nodelist
      nodelist.push_front(omnode(string2int(namelist[n]->d_name), 0, 0));
      howmany++;
      free(namelist[n]);
      }
    }
    free(namelist);

    // adjust clusternodes
    setclusternodes(howmany);

    // check that we are not alone
    // prevents from dividing through zero
    if(howmany>1) {

      // here we calculate the space between the nodes
      nodeangle = 360/(howmany-1);
      noderadialvalue = ((2*pi)/360)*nodeangle;


       // here the dynamic widgets are created
      list<omnode>::iterator p = nodelist.begin();
      while (p != nodelist.end()) {

        omnode iternode = *p;
        int omid = iternode.getnodeid();
        string nodeidtxt=int2string(omid);

        // check that we are not displaying myself
        if(strcmp(nodeidtxt.c_str(), myid.c_str())) {

          mainradialvalue = mainradialvalue + noderadialvalue;
          nodeX = cos(mainradialvalue);
          nodeY = sin(mainradialvalue);
          nodeX = nodeX*(circlewide/2);
          nodeY = nodeY*(circlewide/2);
          nodeX = xcenter+nodeX;
          nodeY = ycenter+nodeY;

          // store the node position
          // here we have to use p to make the changes on the real nodeslist
          p->setposition(int(nodeX), int(nodeY));

          // all the nodes
          QLabel_nodes[nodes]= new nodewidget(QLabel_main,nodeidtxt.c_str());
          QLabel_nodes[nodes]->setGeometry(int(nodeX),int(nodeY),20,20);
          QLabel_nodes[nodes]->setMinimumSize(0,0);
          QLabel_nodes[nodes]->setPixmap(help);
          QLabel_nodes[nodes]->setFrameStyle( QFrame::Panel | QFrame::Sunken );

          QLabel_nodeID[nodes]= new nodewidget(QLabel_main,nodeidtxt.c_str());
          QLabel_nodeID[nodes]->setGeometry(int(nodeX),int(nodeY+20),40,20);
          QLabel_nodeID[nodes]->setMinimumSize(0,0);
          QLabel_nodeID[nodes]->setFont(font);
          QLabel_nodeID[nodes]->setText(nodeidtxt.c_str());
          QLabel_nodeID[nodes]->setFrameStyle( QFrame::Panel | QFrame::Sunken );

          nodes++;

          } else {

          // label myself with the my nodeID
          QLabel_myid->setText(nodeidtxt.c_str());

        } // end of myself-check
        p++;

      } // end of while

      } else {
      // we are alone
      QLabel_myid->setText(myid.c_str());
    }

    QLabel_main->show();
    QTimer *processtimer = new QTimer( QLabel_main, "QLabel_main" );
    connect( processtimer, SIGNAL(timeout()), this, SLOT(fillprocesslist()) );
    processtimer->start(2000);

    fillprocesslist();


  }

}




void OpenMosixmigmonApp::redraw() {
  if(!checknodes()) {
    delete QLabel_main;
    updatenodes();
  }
}


// ############# process handling methods ################

void OpenMosixmigmonApp::fillprocesslist() {

	DIR *dir;
	struct dirent *dir_info;
	string processname;
  string processtip;
	string temp;
	string wherevalue;
  string cmdlinevalue;
	string procdir = "/proc/";
	string stripwhere;
	char zeile[201];
  int pid;
  int home;
  int createnew=0;
  int testpid1=0;
  int testpid2=0;
  int stillthere=0;


  // get a copy from the process-list
  list<oMProc> processlist_copy = processlist;
  processlist.clear();

	if ((dir=opendir("/proc"))!=NULL) {

		while ((dir_info = readdir(dir))!=NULL) {
			processname.erase();
      processtip.erase();
			processname = processname + dir_info->d_name;

			if((strchr(processname.c_str(), '1'))||(strchr(processname.c_str(), '2'))||(strchr(processname.c_str(), '3'))||(strchr(processname.c_str(), '4'))||(strchr(processname.c_str(), '5'))||(strchr(processname.c_str(), '6'))||(strchr(processname.c_str(), '7'))||(strchr(processname.c_str(), '8'))||(strchr(processname.c_str(), '9')) ) {

   			temp.erase();
				temp = temp + procdir;
				temp = temp + processname;
				temp = temp + "/cmdline";				
				std::ifstream readproccmd(temp.c_str());
				if (readproccmd) {
					readproccmd.getline(zeile, 200);
					cmdlinevalue.erase();
					cmdlinevalue= cmdlinevalue+zeile;
					readproccmd.close();
				}

				if (cmdlinevalue.length()>0) {
          cmdlinevalue=onlyfirstpart(cmdlinevalue);
          } else {
          cmdlinevalue="0";
        }

   			temp.erase();
				temp = temp + procdir;
				temp = temp + processname;
				temp = temp + "/where";				
				std::ifstream readprocwhere(temp.c_str());
				if (readprocwhere) {
					readprocwhere.getline(zeile, 200);
					wherevalue.erase();
					wherevalue= wherevalue+zeile;
					readprocwhere.close();
				}


        // insert the process into the processlist
        pid = string2int(processname);
        home = string2int(wherevalue);
        processlist.push_back(oMProc(pid, home, cmdlinevalue));
        processtip = processtip + processname;
        processtip = processtip + ",";
        processtip = processtip + cmdlinevalue;

        // if the "old" processlist has any members
        if (!processlist_copy.size()) {
          singleproc[pid] = new procwidget(QLabel_main,processtip.c_str());

          } else {
          // check if the pid was in the "old" processlist
          createnew=0;
          list<oMProc>::iterator q = processlist_copy.begin();
          while (q != processlist_copy.end()) {
            oMProc oldproc =*q;
            if (oldproc.getompid()==pid) {
              // it was there
              createnew=1;
            }
            q++;
          }
          if(createnew==0) {
            // create it
//            cout << "new processe created" << endl;
            singleproc[pid] = new procwidget(QLabel_main,processtip.c_str());
          }

        }

			}
		} //while
		closedir(dir);
	} //if


  // now check if there are any processes
  // which are not there any more
  list<oMProc>::iterator w = processlist_copy.begin();
  while (w != processlist_copy.end()) {
    oMProc isprocstillthere1 =*w;
    testpid1 = isprocstillthere1.getompid();

    stillthere=0;
    list<oMProc>::iterator o = processlist.begin();
    while (o != processlist.end()) {
      oMProc isprocstillthere2 =*o;
      testpid2 = isprocstillthere2.getompid();

      if (testpid1==testpid2) {
        // it is still there
        stillthere=1;
      }
     o++;
    }
    if(stillthere==0) {
      // it is not there any more we delete -> singleproc instance
//      cout << "a process exits" << endl;
      delete singleproc[testpid1];
    }
    w++;
  }


  // trigger to draw the processlist
  QLabel_main->repaint();
  drawprocesslist();

}




int OpenMosixmigmonApp::gethowmanyprocs() {
 return int(processlist.size());
}



void OpenMosixmigmonApp::drawprocesslist() {

  int howmanyprocs = gethowmanyprocs();
  double procangle=0;
  double procradialvalue=0;
  double mainprocradialvalue=0;
  double procX, procY;
  double remoteX, remoteY;
  int remoteproccount=0;

  double remoteprocangle=0;
  double remoteprocradialvalue=0;
  double remotemainprocradialvalue=0;
  double remprocX, remprocY;
  int migratedprocs=0;
  int whichremoteproc=0;
  int whichsingleproc=0;

  // calculate the processcircle
  procangle = 360/(howmanyprocs);
  procradialvalue = ((2*pi)/360)*procangle;

  // draws the nodecircle
  QPainter ncircle(QLabel_main);
  ncircle.setPen(QPen(Qt::black, 0, SolidLine));
  ncircle.drawEllipse(0, 0, eheight, ewide);

  // here we go through the processlist and draw
  list<oMProc>::iterator p = processlist.begin();
  while (p != processlist.end()) {
    // get a pointer to a proc
    oMProc proc =*p;

    mainprocradialvalue = mainprocradialvalue + procradialvalue;
    procX = cos(mainprocradialvalue);
    procY = sin(mainprocradialvalue);
    procX = procX*(circlewide/4);
    procY = procY*(circlewide/4);
    procX = xcenter+procX;
    procY = ycenter+procY;


    whichsingleproc = proc.getompid();

    if(proc.islocal()) {
    // it is a local proc

      singleproc[whichsingleproc]->setGeometry(int(procX),int(procY),5,5);
      singleproc[whichsingleproc]->setMinimumSize(0,0);
      singleproc[whichsingleproc]->setBackgroundColor(Qt::black);
      singleproc[whichsingleproc]->show();

      } else {

     // it is a remote proc
      // here we have to find an intelligent way to draw the
      // process at the remote node
      // remote node X value
      list<omnode>::iterator rempnode = nodelist.begin();
      while(rempnode != nodelist.end()) {
        omnode realnode =*rempnode;
        if(realnode.getnodeid() == proc.ishere()) {
         remoteX=realnode.getXpos();
         remoteY=realnode.getYpos();
        }
        rempnode++;
      }


      // draw the process at the remote node

      // count howmany remote process are on this node
      list<oMProc>::iterator hmremote = processlist.begin();
      while(hmremote != processlist.end()) {
        oMProc migprocs =*hmremote;
        if(!migprocs.islocal()) {
          migratedprocs++;
        }
        hmremote++;
      }

      // which remote processes is this
      whichremoteproc++;

     // calculate the remote processcircle
     remoteprocangle = 360/migratedprocs+1;
     remoteprocradialvalue = ((2*pi)/360)*remoteprocangle;

     remotemainprocradialvalue = remoteprocradialvalue*whichremoteproc;

     remprocX = cos(remotemainprocradialvalue);
     remprocY = sin(remotemainprocradialvalue);
     remprocX = remprocX*45;
     remprocY = remprocY*45;
     remprocX = remoteX+remprocX+7;
     remprocY = remoteY+remprocY+7;

     // calculate the remote processcircle
     // the paint device for the nodescircles
     QPainter nodescircles(QLabel_main);
     nodescircles.setPen(QPen(Qt::black, 0, SolidLine));
     nodescircles.drawEllipse(int(remoteX-40),int(remoteY-40),100,100);

     // draw a line from the home-position to the remote position
     nodescircles.setPen(QPen(Qt::black, 0, DotLine));
     nodescircles.moveTo(int(procX), int(procY));
     nodescircles.lineTo(int(remprocX), int(remprocY));

     // here we have the data for the remote process circle together
     // and now we draw the remote process
     singleproc[whichsingleproc]->setGeometry(int(remprocX),int(remprocY),5,5);
     singleproc[whichsingleproc]->setMinimumSize(0,0);
     singleproc[whichsingleproc]->setBackgroundColor(Qt::green);
     singleproc[whichsingleproc]->show();

     migratedprocs=0;
     remoteproccount=0;

    } // end of if-remote-or-local

//    singleproc->show();

    p++;
  }


}





void OpenMosixmigmonApp::paintEvent(QPaintEvent *) {

   // this->fillprocesslist();

}







// the help-method
void OpenMosixmigmonApp::slotHelp() {

  QPixmap help, white_pix;
	help = QPixmap(helpicon);
  white_pix = QPixmap(white_xpm);

	QTabDialog *tabs = new QTabDialog();
	tabs->setGeometry(200,50,20,20);
	tabs->resize(300, 600);

	QLabel *a = new QLabel(tabs, 0, 0);
	a->setGeometry(5,5,20,20);
	a->resize(300, 600);
	a->setMargin(3);
	a->setFrameStyle( QFrame::Panel | QFrame::Sunken );
	a->setBackgroundPixmap(QPixmap(white_pix));
	a->setText("openMosixmigmon Help\n\n" \
        "General:\n" \
        "The openMosixmigmon is a monitor for migrations in your openMosix-cluster.\n" \
        "It displays all your nodes as little penguins sitting in a circle.\n" \
        "-> nodes-circle\n" \
        "The main penguin is the node on which openMosixmigmon runs and around\n" \
        "this node it shows its processes also in a circle of small black squares.\n" \
        "-> main process-circle\n" \
        "If a process migrates to one of the nodes the node gets an own process-circle\n" \
        "and the process moved from the main process-circle to the remote process-circle.\n" \
        "Then the process is marked green and draws a line from its origin to its\n" \
        "remote location to visualize the migration.\n" \
        "\n" \
        "\n" \
        "Tooltips:\n" \
        "If you hold your mouse above a process it will show you its PID and commandline\n" \
        "in a small tooltip-window.\n" \
        "\n" \
        "\n" \
        "Drag'n Drop!\n" \
        "The openMosixmigmon is fully Drag'n Drop enabled.\n" \
        "You can grab (drag) any process and drop them to any of your nodes (those penguins)\n" \
        "and the process will move there.\n" \
        "If you double-click a process on a remote node it will be send home immediately.\n" \
        "\n" \
        "\n" \
        "I hope you have much fun with it ;)\n" \
        "\n" \
        "\nM. Rechenburg / mosixview@t-online.de");


	tabs->addTab(a, help,  "general");
	tabs->show();
}






